      /////////////////////////////////////////////
     // THE CONSTRAINT-UNDERSTANDING SYSTEM      //
    //  (Version 2)                             ///
   // Written by Andrew Broad                  ////
  //                                          /////
 // ExtractionEngine.java (2001:11:15 16:56) //////
///////////////////////////////////////////////////
// Methods of ExtractionEngine:             ///////
// - extraction                             //////
// - CCUSinit                               /////
// - CCUSextract                            ////
// - submitForNavigation                    ///
// - submitForExtraction                    //
/////////////////////////////////////////////

import java.util.*;
import java.io.*;

import jex.*;

/*****************************************************************************
** The class `ExtractionEngine' contains the procedure that drives the 
** extraction of higher-level knowledge from code. It uses meta cases to
** navigate the source code in a data-driven manner.
** This class is not meant to be instantiated.
*****************************************************************************/

public class ExtractionEngine {
	public static boolean _prettyPrintHLCs= false;
	public static boolean _showIntermediateOutput= true;
	public static int _numberOfHLCsExtracted;

	private static Vector/*Frame*/ _extractionQueue= new Vector();
	private static Hashtable/*Frame->Boolean*/ _alreadyPromptedExtraction
	                                           = new OrderedHashtable();
	private static Hashtable/*Frame->Boolean*/ _alreadyPromptedNavigation
	                                           = new OrderedHashtable();
	private static OrderedHashtable/*FrameClass->Vector<Frame>*/
	               _HLCsummary= new OrderedHashtable();
	private static CaseLibrary _extractionCaseLibrary;
	private static CaseLibrary _navigationCaseLibrary;
	private static PrettyPrint _prettyPrint;

	/********************************************************************/
	// static ExtractionEngine.extraction : (Frame x CaseLibrary x
	//                                       CaseLibrary x CaseLibrary)
	//                                      -> void
	/*********************************************************************
	** Extracts the higher-level knowledge from a piece of code encoded
	** as a frame network.
	** @param topLevelFrame The top-level frame in the frame network
	**                      representing the input code.
	** @param extractionCaseLibrary The case library containing the
	**                              extraction cases.
	** @param navigationCaseLibrary The case library containing the
	**                              navigation cases.
	** @param prettyPrintCaseLibrary The case library containing the
	**                               pretty-print cases.
	*********************************************************************/

	public static void extraction (Frame topLevelFrame,
	                               CaseLibrary extractionCaseLibrary,
	                               CaseLibrary navigationCaseLibrary,
	                               CaseLibrary prettyPrintCaseLibrary)
	{
		_extractionCaseLibrary= extractionCaseLibrary;
		_navigationCaseLibrary= navigationCaseLibrary;
		_prettyPrint= new PrettyPrint(prettyPrintCaseLibrary);

		//////////////////////////////////////////////////////
		// Walk the code, extracting higher-level knowledge //
		//////////////////////////////////////////////////////

		submitForNavigation(topLevelFrame);

		/////////////////////
		// Produce summary //
		/////////////////////

		System.out.println("----------------------------------------"
			        + "----------------------------------------");
		System.out.println("SUMMARY\n-------\n");
		System.out.println("The following higher-level knowledge was "
		+ "extracted:");

		ListIterator HLCs= _HLCsummary.keyListIterator();
		while (HLCs.hasNext()) {
			FrameClass key= (FrameClass)HLCs.next();
			Vector v= (Vector)_HLCsummary.get(key);
			System.out.print("\n" + v.size() +" "+ key + " HLC");
			if (v.size() > 1) System.out.print("s");

			if (_prettyPrintHLCs) {
				System.out.println(":-");
			}

			Enumeration vHLCs= v.elements();
			while (vHLCs.hasMoreElements()) {
				Frame HLC= (Frame)vHLCs.nextElement();

				if (_prettyPrintHLCs) {
					System.out.print("\t" + HLC);
				}

				String description= 
				  _prettyPrint.generateDescription(HLC,true);

				if (_prettyPrintHLCs) {
					System.out.print(": " + description);
					System.out.println();
				}
			}
		}
		System.out.println();
		_numberOfHLCsExtracted= _extractionQueue.size();
	}

	/********************************************************************/
	// static ExtractionEngine.CCUSinit : (CaseLibrary x CaseLibrary x
	//                                     CaseLibrary)
	//                                    -> void
	/*********************************************************************
	** Extracts the higher-level knowledge from a piece of code encoded
	** as a frame network.
	** @param topLevelFrame The top-level frame in the frame network
	**                      representing the input code.
	** @param extractionCaseLibrary The case library containing the
	**                              extraction cases.
	** @param navigationCaseLibrary The case library containing the
	**                              navigation cases.
	** @param prettyPrintCaseLibrary The case library containing the
	**                               pretty-print cases.
	*********************************************************************/

	public static void CCUSinit (CaseLibrary extractionCaseLibrary,
	                             CaseLibrary navigationCaseLibrary,
	                             CaseLibrary prettyPrintCaseLibrary)
	{
		_extractionCaseLibrary= extractionCaseLibrary;
		_navigationCaseLibrary= navigationCaseLibrary;
		_prettyPrint= new PrettyPrint(prettyPrintCaseLibrary);
	}

	/********************************************************************/
	// static ExtractionEngine.CCUSextract : Frame -> OrderedHashtable
	/*********************************************************************
	** An interface to the extraction engine for the HLC-extracting CCUS.
	** @param promptFrame The prompt frame for extraction.
	** @return A hash table containing the HLCs extracted from the given
	**         prompt.
	*********************************************************************/

	public static OrderedHashtable CCUSextract (Frame promptFrame) {
		_HLCsummary= new OrderedHashtable();
		_extractionQueue= new Vector();
		_numberOfHLCsExtracted= 0;

		//////////////////////////////////////////////////////
		// Walk the code, extracting higher-level knowledge //
		//////////////////////////////////////////////////////

		submitForNavigation(promptFrame);
		submitForExtraction(promptFrame);

		/////////////////////
		// Produce summary //
		/////////////////////

		ListIterator HLCs= _HLCsummary.keyListIterator();
		while (HLCs.hasNext()) {
			FrameClass key= (FrameClass)HLCs.next();
			Vector v= (Vector)_HLCsummary.get(key);

			Enumeration vHLCs= v.elements();
			while (vHLCs.hasMoreElements()) {
				Frame HLC= (Frame)vHLCs.nextElement();

				CCUS_FMA.indent();
				System.out.print("+ " + HLC);

				String description= 
				  _prettyPrint.generateDescription(HLC,true);

				if (_prettyPrintHLCs) {
					System.out.print(": " + description);
				}
				System.out.println();
			}
		}
		_numberOfHLCsExtracted= _extractionQueue.size();
		return _HLCsummary;
	}

	/********************************************************************/
	// static ExtractionEngine.submitForNavigation : Frame -> void
	/*********************************************************************
	** Uses the given frame as a prompt to retrieve a navigation case,
	** which suggests what frames to submit as extraction prompts.
	** @param promptFrame The prompt frame.
	*********************************************************************/

	private static void submitForNavigation (Frame promptFrame) {

		/////////////////////////////////////////////////
		// Don't use any frame as an navigation prompt //
		// more than once.                             //
		/////////////////////////////////////////////////
		if (_alreadyPromptedNavigation.get(promptFrame) != null)
			return;
		else
			_alreadyPromptedNavigation.
			put(promptFrame, new Boolean(true));

		if (_showIntermediateOutput) {
			System.out.println("Submitting " + promptFrame
			+ " as a NAVIGATION prompt.");
		}

		Vector navigationCases= _navigationCaseLibrary.
		                        retrieve(promptFrame);

		if (_showIntermediateOutput) {
			System.out.println("ExtractionEngine.submitForNavigat"
			+ "ion(" + promptFrame + "): " + navigationCases.size(
			) + " navigation cases retrieved");
		}

		if (navigationCases.size() >= 1) {
			Case navigationCase
			= CaseLibrary.selectMostSpecific(navigationCases);

			if (_showIntermediateOutput) {
				System.out.print("Matching " + navigationCase.
				  getCaseFrame() + " - ");
			}

			if (navigationCase.match(promptFrame)) {
				Vector submits= navigationCase.extract();
				if (submits.size() > 1) {
					System.err.println("You'll have to ex"
					+ "tend ExtractionEngine.submitForNav"
					+ "igation if you want to have multip"
					+ "le solution frames in navigation c"
					+ "ases! ;-)");
					System.exit(1);
				}
				Frame submit= (Frame)submits.elementAt(0);

				if (_showIntermediateOutput) {
					System.out.println
					           ("match succeeded.");
				}

				Vector submissionFrames
				= (Vector)submit.gsf("frames");
				Enumeration sf= submissionFrames.elements();
				Object item;
				while (sf.hasMoreElements()) {
					item= sf.nextElement();
					if (item instanceof Frame) {
						/////////////////////////////
						// process subexpressions  //
						// before superexpressions //
						// (CHANGE BELOW TOO)      //
						/////////////////////////////
						submitForNavigation
						  ((Frame)item);
						submitForExtraction
						  ((Frame)item);
					} else if (item instanceof Vector) {
						Enumeration f
						= ((Vector)item).elements();
						while (f.hasMoreElements()){
							Object frame=
							  f.nextElement();
							if (frame instanceof
							          Frame)
							{
							   submitForNavigation
							       ((Frame)frame);
							   submitForExtraction
							       ((Frame)frame);
							}
							else
							   System.err.println
							   ("ExtractionEngine"
							   +".submitForNaviga"
							   + "tion: Warning: "
							   + frame + " is not"
							   + " a Frame - not "
							   + "submitted");
						}
					}
				}
				Frame.remove(submit);
			} else {
				System.err.println("ExtractionEngine.extrac"
				+ "tion: navigation case failed to match!");
				System.exit(1);
			}
		}

		//////////////////////////////////////////////////////
		// Walk the HLCs already extracted from the entity, //
		// extracting further HLCs.                         //
		//////////////////////////////////////////////////////

		for (int i= 0; i < _extractionQueue.size(); i++) {
			submitForExtraction
			  ((Frame)_extractionQueue.elementAt(i));
		}
	}

	/********************************************************************/
	// static ExtractionEngine.submitForExtraction : Frame -> void
	/*********************************************************************
	** Uses the given frame as a prompt to retrieve an extraction case,
	** match it, and extract the HLC it suggests if it does match.
	** @param promptFrame The prompt frame.
	*********************************************************************/

	private static void submitForExtraction (Frame promptFrame) {

		/////////////////////////////////////////////////
		// Don't use any frame as an extraction prompt //
		// more than once.                             //
		/////////////////////////////////////////////////
		if (_alreadyPromptedExtraction.get(promptFrame) != null)
			return;
		else
			_alreadyPromptedExtraction.
			put(promptFrame, new Boolean(true));

		if (_showIntermediateOutput) {
			System.out.print("+------------");
			for (int c= 0; c<promptFrame.toString().length(); c++)
				System.out.print("-");
			System.out.println("-------------------------+");
			System.out.println("| Submitting " + promptFrame
			+ " as an EXTRACTION prompt |");
			System.out.print("+------------");
			for (int c= 0; c<promptFrame.toString().length(); c++)
				System.out.print("-");
			System.out.println("-------------------------+");
		}

		//////////////////////////////////////////
		// Retrieve potentially matching cases. //
		//////////////////////////////////////////
		Vector cases= _extractionCaseLibrary.retrieve(promptFrame);

		////////////////////////////////////////////
		// Match cases to prompt, extracting HLCs //
		// when matches succeed.                  //
		////////////////////////////////////////////
		if (_showIntermediateOutput) {
			System.out.println("Matching it with the following "
			+ "cases...");
		}

		Enumeration casesEnumeration= cases.elements();
		while (casesEnumeration.hasMoreElements()) {
			Case c= (Case)casesEnumeration.nextElement();

			if (_showIntermediateOutput) {
				System.out.print("\t" + c.getCaseFrame()
				+ " - ");
			}

			if (c.match(promptFrame)) {
				Enumeration HLCsExtracted
				= (c.extract()).elements();
				while (HLCsExtracted.hasMoreElements()) {
					Frame HLC
					= (Frame)HLCsExtracted.nextElement();

					if (_showIntermediateOutput) {
						System.out.println("match suc"
						+ "ceeded, " + HLC
						+ " extracted.");
					}

					_extractionQueue.addElement(HLC);
					FrameClass fc= HLC.getFrameClass();
					Vector v= (Vector)_HLCsummary.get(fc);
					if (v == null) v= new Vector();
					v.addElement(HLC);
					_HLCsummary.put(fc,v);
				}
			}
		}
		////////////////////////////////////////////////////
		// Extracting implicit conjunctions from entities //
		////////////////////////////////////////////////////
		if(promptFrame.getFrameClass().getName().equals("entity_type")
		&& promptFrame.gsf("domain_rules") != null
		&& ((Vector)promptFrame.gsf("domain_rules")).size() >= 2)
		{
			Frame conjunction= new Frame("conjunction");
			UnorderedVector conjuncts= new UnorderedVector();
			ListIterator domainRules= ((Vector)promptFrame.
				gsf("domain_rules")).listIterator();
			while (domainRules.hasNext()) {
				conjuncts.addElement(((Frame)
				  domainRules.next()).gsf("expression"));
			}
			conjunction.ssf("conjuncts", conjuncts);
			conjunction.ssf("extracted_from", promptFrame);

			if (_showIntermediateOutput) {
				System.out.println(conjunction
				+ " extracted from " + promptFrame + ".");
			}

			FrameClass fc= FrameClass.lookup("conjunction");
			Vector v= (Vector)_HLCsummary.get(fc);
			if (v == null) v= new Vector();
			v.addElement(conjunction);
			_HLCsummary.put(fc,v);

			if (_showIntermediateOutput) {
				System.out.println();
			}

			submitForExtraction(conjunction);
		}		
		if (_showIntermediateOutput) {
			System.out.println();
		}
	}
}
