      ////////////////////////////////////////
     // THE CONSTRAINT-UNDERSTANDING SYSTEM //
    //  (Version 2)                        ///
   // Written by Andrew Broad             ////
  //                                     /////
 // FrameClass.java (2001:09:20 14:49)  //////
//////////////////////////////////////////////
// Methods of FrameClass:              ///////
// - FrameClass constructor            ///////
// - getClassFrame                     ///////
// - getName                           ///////
// - isAbstract                        ///////
// - getSubClasses                     ///////
// - getSuperClasses                   ///////
// - subsumes                          ///////
// - getSlots                          ///////
// - getInstances                      ///////
// - addInstance                       ///////
// - removeInstance                    ///////
// - toString                          ///////
// - getSlot                           ///////
// - generateFILforInstances           ///////
// - static createFrameClasses         ///////
// - static getAllFrameClasses         ///////
// - static lookup(String)             ///////
// - static lookupX(String)            //////
// - static lookup(String,Model)       /////
// - static lookup(Entity,Model)       ////
// - static writeClass                 ///
// - static readClass                  //
////////////////////////////////////////

import java.io.*;
import java.util.*;

import jex.*;

/*****************************************************************************
** The class `FrameClass' represents a class of frames, serving to classify
** them and to specify the slots they should have.
*****************************************************************************/

public class FrameClass implements Serializable {
	private Entity _underlyingEntity;
	private String _name;
	private Vector/*Frame*/ _instances= new Vector();
	private boolean _anyInputModelFrames= false;
	private Vector/*FrameClass*/ _superClasses= new Vector();
	private Vector/*FrameClass*/ _subClasses= new Vector();
	private Vector/*Slot*/ _slots= new Vector();
	private Hashtable/*String,Slot*/ _slotIndex= new Hashtable();
	private static Hashtable/*String,FrameClass*/ _frameClassIndex
	                                              = new Hashtable();

	/********************************************************************/
	// FrameClass constructor : Entity x Schema ->
	/*********************************************************************
	** Constructs this frame class by plugging in the given Jex Entity
	** object which is taken to represent it (it does so by copying the
	** reference, which is maximally efficient, but the client should not
	** then alter the Entity directly, as that would compromise the
	** integrity of the frame).
	** @param fcEntity A Jex Entity object to wrap in this frame.
	** @param frameClassSchema The EXPRESS schema in which this
	**  frame class is defined, in case it has any superclasses which
	**  haven't been constructed yet.
	*********************************************************************/

	public FrameClass (Entity fcEntity, Schema frameClassSchema) {
		Slot slot; FrameClass frameClass;

		_underlyingEntity= fcEntity;

		try {
			_name= fcEntity.getId();	
		}
		catch (UnsetValueException uve) {
			System.err.println("UnsetValueException caught in "
			+ "FrameClass constructor");
			System.exit(1);
		}
		if (_frameClassIndex.get(_name) != null) {
			System.err.println("FrameClass \"" + _name + "\" "
			+ "already exists!");
			System.exit(1);
		}

		_frameClassIndex.put(_name,this);

		////////////////////////////////////////
		// Construct the slots inherited from //
		// this frame class's superclasses    //
		////////////////////////////////////////

		ListIterator supertypes= fcEntity.getEntityRefs();
		while (supertypes.hasNext()) {
			String stypeName= null;
			try {
				stypeName= ((Ref)supertypes.next()).getId();
			} catch (UnsetValueException uve) {
				System.err.println("UnsetValueException "
				+ "caught in FrameClass constructor");
				System.exit(1);
			}
			frameClass= null;
			try {
				frameClass= lookup(stypeName,frameClassSchema);
				_superClasses.addElement(frameClass);
				frameClass.getSubClasses().addElement(this);
				Vector inheritedSlots= frameClass.getSlots();
				for (int a= 0; a < inheritedSlots.size(); a++)
				{
					slot= (Slot)
					      inheritedSlots.elementAt(a);

					if (_slotIndex.
					    get(slot.getName()) == null)
					{
						_slots.addElement(slot);
					}
					_slotIndex.put(slot.getName(),slot);
				}
			} catch (FrameClassDoesNotExist x) {
				System.err.println("Something has gone very "
				+ "wrong for the frame class " + stypeName
				+ " not to exist!");
			}
		}

		/////////////////////////////////////////////////
		// Construct this frame class's explicit slots //
		/////////////////////////////////////////////////

		ListIterator explicits= fcEntity.getExplicitAttributes();
		while (explicits.hasNext()) {
			ExplicitAttribute explicit
			= (ExplicitAttribute)explicits.next();

			slot= new Slot(explicit,this,frameClassSchema);
			_slots.addElement(slot);
			try {
				_slotIndex.put(explicit.getId(),slot);
			} catch (UnsetValueException uve) {
				System.err.println("UnsetValueException "
				+ "caught in FrameClass constructor");
				System.exit(1);
			} 
		}

		////////////////////////////////////////////////
		// Construct this frame class's derived slots //
		////////////////////////////////////////////////

		ListIterator deriveds= fcEntity.getDerivedAttributes();
		while (deriveds.hasNext()) {
			DerivedAttribute derived
			= (DerivedAttribute)deriveds.next();

			slot= new Slot(derived,this,frameClassSchema);
			_slots.addElement(slot);
			try {
				_slotIndex.put(derived.getId(),slot);
			} catch (UnsetValueException uve) {
				System.err.println("UnsetValueException "
				+ "caught in FrameClass constructor");
				System.exit(1);
			} 
		}

		////////////////////////////////////////////////
		// Construct this frame class's inverse slots //
		////////////////////////////////////////////////

		ListIterator inverses= fcEntity.getInverseAttributes();
		while (inverses.hasNext()) {
			InverseAttribute inverse
			= (InverseAttribute)inverses.next();

			slot= new Slot(inverse,this,frameClassSchema);
			_slots.addElement(slot);
			try {
				_slotIndex.put(inverse.getId(),slot);
			} catch (UnsetValueException uve) {
				System.err.println("UnsetValueException "
				+ "caught in FrameClass constructor");
				System.exit(1);
			} 
		}
	}

	/********************************************************************/
	// FrameClass.getClassFrame : void -> Frame
	/*********************************************************************
	** Delivers up the underlying frame of this frame class.
	** @return The class frame representing this frame class.
	*********************************************************************/

	public Frame getClassFrame () {
		return JexToFrames.getCachedFrame(_underlyingEntity);
	}

	/********************************************************************/
	// FrameClass.getName : void -> String
	/*********************************************************************
	** Returns the name of this frame class.
	** @return The name of this frame class.
	*********************************************************************/

	public String getName () {
		return _name;
	}

	/********************************************************************/
	// FrameClass.isAbstract : void -> boolean
	/*********************************************************************
	** Returns true just in case this frame class is abstract (i.e. cannot
	** be directly instantiated).
	** @return Whether this frame class is abstract.
	*********************************************************************/

	public boolean isAbstract () {
		return _underlyingEntity.isAbstract();
	}

	/********************************************************************/
	// FrameClass.getSubClasses : void -> Vector<FrameClass>
	/*********************************************************************
	** Returns the subclasses of this frame class.
	** @return The subclasses of this frame class.
	*********************************************************************/

	public Vector getSubClasses () {
		return _subClasses;
	}

	/********************************************************************/
	// FrameClass.getSuperClasses : void -> Vector<FrameClass>
	/*********************************************************************
	** Returns the superclasses of this frame class.
	** @return The superclasses of this frame class.
	*********************************************************************/

	public Vector getSuperClasses () {
		return _superClasses;
	}

	/********************************************************************/
	// FrameClass.subsumes : FrameClass -> boolean
	/*********************************************************************
	** Does this frame class subsume another? Returns true just in case
	** the given frame class is equal to or a subclass of this one.
	** @param subClass The frame class to test whether it is a subclass of
	**  this one.
	** @return Whether this frame class does subsume `subClass'.
	*********************************************************************/

	public boolean subsumes (FrameClass subClass) {
		if (subClass == this)
			return true;
		
		//////////////////////////////////////////////////
		// Move up the hierarchy from subClass          //
		// and see if you collide with this frame class //
		//////////////////////////////////////////////////
		for (int s= 0; s < subClass.getSuperClasses().size(); s++) {
			if (subsumes((FrameClass)subClass.getSuperClasses().
			elementAt(s))) {
				return true;
			}
		}
		return false;
	}

	/********************************************************************/
	// FrameClass.getSlots : void -> Vector<Slot>
	/*********************************************************************
	** Delivers up the slots in this frame.
	** @return A vector of the slots in this frame.
	*********************************************************************/

	public Vector getSlots () {
		return _slots;
	}

	/********************************************************************/
	// FrameClass.getInstances : void -> Vector<Frame>
	/*********************************************************************
	** Returns the instances of this frame class that have been
	** created in the program's lifetime.
	** @return A vector of frames that are instances of this frame class.
	*********************************************************************/

	public Vector getInstances () {
		return _instances;
	}

	/********************************************************************/
	// FrameClass.anyInputModelFrames : void -> boolean
	/*********************************************************************
	** Returns true just in case this frame class has any instance frames
	** from the actual input model.
	** @return Whether this frame class has any input model frames.
	*********************************************************************/

	public boolean anyInputModelFrames () {
		return _anyInputModelFrames;
	}

	/********************************************************************/
	// FrameClass.addInstance : Frame -> void
	/*********************************************************************
	** Adds the given frame to the list of instances of this FrameClass.
	** @param frame The instance frame to add (must be of this frame
	**              class).
	** @exception InvalidFrameClass If the given frame is not of this
	**                              frame class!
	*********************************************************************/

	public void addInstance (Frame frame) throws InvalidFrameClass {
		if (frame.getFrameClass() != this)
			throw new InvalidFrameClass();

		_instances.addElement(frame);

		if (frame.isInputModelFrame())
			_anyInputModelFrames= true;
	}

	/********************************************************************/
	// FrameClass.removeInstance : Frame -> void
	/*********************************************************************
	** Removes the given frame from the list of instances of this
	** FrameClass.
	** @param frame The instance frame to remove (must be of this frame
	**              class).
	** @exception InvalidFrameClass If the given frame is not of this
	**                              frame class!
	*********************************************************************/

	public void removeInstance (Frame frame) throws InvalidFrameClass {
		if (frame.getFrameClass() != this)
			throw new InvalidFrameClass();

		_instances.remove(frame);

		if (frame.isInputModelFrame()) {
			_anyInputModelFrames= false;
			ListIterator instances= _instances.listIterator();
			while (instances.hasNext()) {
				if (((Frame)instances.next()).
				    isInputModelFrame())
				{
					_anyInputModelFrames= true;
				}
			}
		}
	}

	/********************************************************************/
	// FrameClass.toString : void -> String
	/*********************************************************************
	** Returns a string representation of this frame class.
	** @return The string representation of this frame class.
	*********************************************************************/

	public String toString () {
		return _name;
	}

	/********************************************************************/
	// FrameClass.getSlot : String -> Slot
	/*********************************************************************
	** Looks up the slot with the given name.
	** @param slotName The name of the slot to return.
	** @return The Slot object whose name is `slotName'.
	** @exception InvalidSlot If the frame class does not have a slot of
	**                        that name.
	*********************************************************************/

	public Slot getSlot (String slotName)
	                    throws InvalidSlot
	{
		Slot slot= (Slot)_slotIndex.get(slotName);
		if (slot == null)
			throw new InvalidSlot();
		return slot;
	}

	/********************************************************************/
	// Frame.generateFILforInstances : boolean -> void
	/*********************************************************************
	** Writes all instances of this frame class into a FIL file, called
	** <TT><<I>frame-class</I>>.fil</TT>.
	** @param HTML If true, the file will be augmented with HTML tags and
	**             have the extension <TT>.html</TT> instead of
	**             <TT>.fil</TT>.
	** @param suppress If true, only input model frames will be included
	**                 in the file (i.e. frame class model frames and
	**                 frames that are part of cases will be excluded).
	** @param indexHTML The PrintWriter for index.html (optional).
	*********************************************************************/

	public void generateFILforInstances (boolean HTML, boolean suppress,
	                                     PrintWriter indexHTML)
	{
		try {
			String filename= _name;
			if (HTML) filename+= ".html";
			else      filename+= ".fil";

			System.out.println("Generating " + filename);
			FileOutputStream fo= new FileOutputStream
			                       ("/usr/local/tmp/users/broada/"
			                       + "FRAMES/" + filename);
			PrintWriter pw= new PrintWriter(fo,true);

			if (HTML) pw.println("<HTML>");
			if (HTML) pw.println("<TITLE>" + _name + ".html"
			                    + "</TITLE>");
			if (HTML) pw.println("<H1>" + _name + "</H1>");

			Enumeration instanceFrames= getInstances().elements();
			while (instanceFrames.hasMoreElements()) {
				Frame frame= (Frame)instanceFrames.
				             nextElement();
				if (!suppress || frame.isInputModelFrame()) {
					if (HTML) pw.println("<HR>");
					else      pw.println();
		       			frame.generateFIL(pw,HTML,suppress);
		       		}
			}
			if (HTML) pw.println("</HTML>");
			fo.close();

			if (indexHTML != null) {
				indexHTML.println("<LI><A HREF=\"" + filename
				+ "\"><TT>" + _name + "</TT></A>");
			}
		} catch (IOException y) {
			System.err.println(y);
			System.exit(1);
		}
	}

	/********************************************************************/
	// static FrameClass.createFrameClasses : Schema -> void
	/*********************************************************************
	** Create a FrameClass for every entity in a given schema.
	** @param schema The schema.
	*********************************************************************/

	public static void createFrameClasses (Schema schema) {
		ListIterator entities= schema.getEntitys();
		while (entities.hasNext()) {
			Entity entity= (Entity)entities.next();
			try {
				if (_frameClassIndex.get(entity.getId())
				    == null)
				{
					new FrameClass(entity,schema);
				}
			} catch (UnsetValueException uve) {
				System.err.println("FrameClass.create"
				+ "FrameClasses: UnsetValueException");
				System.exit(1);
			}
		}
	}

	/********************************************************************/
	// static FrameClass.getAllFrameClasses : void -> Enumeration
	/*********************************************************************
	** Returns all the frame classes in the system.
	** @return An enumeration of all the frame classes in the system.
	*********************************************************************/

	public static Enumeration getAllFrameClasses () {
		return _frameClassIndex.elements();
	}

	/********************************************************************/
	// static FrameClass.lookup : String -> FrameClass
	/*********************************************************************
	** Looks up the frame class with the given name. A quick and
	** convenient method which assumes that the frame class of that name
	** (already) exists - if not, it will report an error and exit from
	** within.
	** @param frameClassName The name of the frame class to return.
	** @return The FrameClass object whose name is `frameClassName'.
	** @see lookup(String,Schema)
	** @see lookupX(String)
	*********************************************************************/

	public static FrameClass lookup (String frameClassName) {
		FrameClass frameClass= (FrameClass)_frameClassIndex.
		                       get(frameClassName);
		if (frameClass == null) {
			System.err.println("Frame class \"" + frameClassName
			+ "\" does not exist [tip: use FrameClass.lookup"
			+ "(String,Schema) for greater flexibility than "
			+ "FrameClass.lookup(String)]");
			System.exit(1);
		}
		return frameClass;
	}

	/********************************************************************/
	// static FrameClass.lookupX : String -> FrameClass
	/*********************************************************************
	** Looks up the frame class with the given name, throwing an exception
	** if no such frame class exists.
	** @param frameClassName The name of the frame class to return.
	** @return The FrameClass object whose name is `frameClassName'.
	** @exception FrameClassDoesNotExist If no frame class of that name
	**            exists already.
	** @see lookup(String)
	** @see lookup(String,Schema)
	*********************************************************************/

	public static FrameClass lookupX (String frameClassName)
	throws FrameClassDoesNotExist
	{
		FrameClass frameClass= (FrameClass)_frameClassIndex.
		                       get(frameClassName);
		if (frameClass == null) {
			throw new FrameClassDoesNotExist();
		}
		return frameClass;
	}

	/********************************************************************/
	// static FrameClass.lookup : String x Schema -> FrameClass
	/*********************************************************************
	** Looks up the frame class with the given name. A more cumbersome
	** version of the lookup method, for use when the frame class of that
	** name may not have been constructed yet. If the frame class does not
	** exist, it tries to construct it, looking up the entity of that name
	** in the given EXPRESS schema. Failing that, it throws a
	** FrameClassDoesNotExist exception.
	** @param frameClassName The name of the frame class to return.
	** @param frameClassSchema The EXPRESS model in which `frameClassName'
	**  is defined, in case that frame class hasn't been constructed yet.
	** @return The FrameClass object whose name is `frameClassName'.
	** @exception FrameClassDoesNotExist If no frame class of that name
	**  exists already, and there is no entity of that name in the frame
	**  class schema either.
	** @see lookup(String)
	*********************************************************************/

	public static FrameClass lookup (String frameClassName,
	                                 Schema frameClassSchema)
	throws FrameClassDoesNotExist
	{
		FrameClass frameClass= (FrameClass)_frameClassIndex.
			get(frameClassName);
		if (frameClass == null) {
			Entity e= null;
			try {
				e=frameClassSchema.findEntity(frameClassName);
			}
			catch (NullParameterException npe) {
				System.err.println("NullParameterException ca"
				+ "ught in FrameClass.lookup(String,Schema)");
				System.exit(1);
			}
	      		if (e != null)
				frameClass=new FrameClass(e,frameClassSchema);
			else
				throw new FrameClassDoesNotExist();
		}
		return frameClass;
	}

	/********************************************************************/
	// static FrameClass.lookup : Entity x Schema -> FrameClass
	/*********************************************************************
	** Looks up the frame class corresponding to the given entity,
	** creating such a frame class if it doesn't exist yet.
	** @param frameClassEntity The entity of the frame class to return.
	** @param frameClassSchema The EXPRESS model in which `frameClassName'
	**  is defined, in case that frame class hasn't been constructed yet.
	** @return The FrameClass object constructed from `frameClassEntity'.
	*********************************************************************/

	public static FrameClass lookup (Entity frameClassEntity,
	                                 Schema frameClassSchema)
	{
		FrameClass frameClass= null;
 		try {
			frameClass= (FrameClass)_frameClassIndex.
			            get(frameClassEntity.getId());
		}
		catch (UnsetValueException uve) {
			System.err.println("UnsetValueException caught in "
			+ "FrameClass.lookup");
			System.exit(1);
		}
		if (frameClass == null)
       			frameClass= new FrameClass(frameClassEntity,
       			                           frameClassSchema);
		return frameClass;
	}

	/********************************************************************/
	// static FrameClass.writeClass : ObjectOutput -> void
	/*********************************************************************
	** Serialises the static data members (class variables) of FrameClass.
	** @param oo An output stream to save the class data to.
	** @exception IOException If an I/O exception occurs.
	*********************************************************************/

	public static void writeClass (ObjectOutput oo)
	                              throws IOException
	{
		oo.writeObject(_frameClassIndex);
	}

	/********************************************************************/
	// static FrameClass.readClass : ObjectInputStream -> void
	/*********************************************************************
	** Deserialises the static data members (class variables) of
	** FrameClass.
	** @param oi An input stream to save the class data to.
	** @exception IOException If an I/O exception occurs.
	*********************************************************************/

	public static void readClass (ObjectInput oi)
	                             throws IOException
	{
		try {
			_frameClassIndex= (Hashtable)oi.readObject();
		} catch (ClassNotFoundException x) {
			System.err.println("This can never happen, but the st"
			+ "upid compiler is just too dumb to realise this!");
		}
	}
}
